/// @author: jvrodrigo

import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:crypto/crypto.dart';
import 'package:dcoupon_sdk_flutter_module/constants.dart';
import 'package:dcoupon_sdk_flutter_module/services.dart';
import 'package:intl/intl.dart';

abstract class HttpInterface {
  static const String HTTP_METHOD = "httpMethod";
  static const String URL = "url";
  static const String CLIENT_DOMAIN = "clientDomain";
  static const String PROPERTIES = "properties";
  static const String PAYLOAD = "payload";
  static const String POST = "POST";
  static const String GET = "GET";
  static const String STATUS_CODE = "statusCode";
  static const String RESPONSE_CODE = "responseCode";
  static const String HEADERS = "headers";
  static const String BODY = "body";
  static const String RESPONSE = "response";
  static const String HTTP_MESSAGE = "http-message";
  static const String RESPONSE_MSG = "responseMsg";
  static const String MESSAGE = "message";
  static const String SESSION_TOKEN = "sessionToken";
  static const String JSON_RESPONSE = "jsonResponse";
  static const String CLIENT_TYPE = "client-type";
  static const String LOCATION = "location";
  static const String UTF_8 = "UTF-8";
  static const int TIME_OUT_SC = 15;
  static const int HTTP_INTERNAL_ERROR = 500;
  static const int HTTP_UNAVAILABLE = 503;
  static const int HTTP_UNAUTHORIZED = 401;
  static const int HTTP_NOT_FOUND = 404;
  static const int HTTP_REQUEST_TIMEOUT = 408;
  static const int HTTP_MOVED_PERM = 301;
  static const int HTTP_MOVED_TEMP = 302;
  static const int HTTP_OK = 200;
  static const String BEARER_ = "Bearer ";
  static const String AUTHORIZATION = "Authorization";
  static const String HEADER_AUTH_APIKEY = "dcoupon-authorization-apikey";
  static const String HEADER_AUTH_SIGNATURE = "dcoupon-authorization-signature";
  static const String HEADER_AUTH_METHOD = "dcoupon-authorization-method";
  static const String HEADER_AUTH_TIMESTAMP = "dcoupon-authorization-timestamp";
  static const String SIGNATURE = "SIGNATURE";
  static const String JWT = "JWT";

  static const dynamic URLS_API_COUPONS = {
    'local': 'http://10.0.2.2:3000/',
    'test': 'http://devmpromo.scanlife.com/',
    'dev': 'https://devapi2coupons.dcoupon.com/',
    'dcoupon.eu': 'https://api2coupons.dcoupon.eu/',
    'dcoupon.com': 'https://api2coupons.dcoupon.com/'
  };
  static const dynamic URLS_API_THIRD_PARTY_LOGIN = {
    'local': 'http://10.0.2.2:3000/',
    'test': 'https://0me53fpvz5.execute-api.eu-west-1.amazonaws.com/test/',
    'dev': 'https://5ynjz2t566.execute-api.us-east-1.amazonaws.com/dev/',
    'dcoupon.eu': 'https://dcoupon.eu/',
    'dcoupon.com': 'https://dcoupon.com/'
  };
  static const dynamic URLS_API_PUBLISHER = {
    'local': 'http://10.0.2.2:3000/',
    'test': 'https://testapi.dcoupon.com/',
    'dev': 'https://services-dev.dcoupon.com/',
    'dcoupon.eu': 'https://dcoupon.eu/',
    'dcoupon.com': 'https://dcoupon.com/'
  };

// 3rdPartyLoginAPI services
  static const String THIRD_PARTY_API_VERSION = "v1";
  static const String PATH_THIRD_PARTY_LOGIN = "login-sdk/" + THIRD_PARTY_API_VERSION;
  static const String PATH_THIRD_PARTY_ACCEPT_ENROLLMENT = "acceptenrollment-sdk/" + THIRD_PARTY_API_VERSION;
  static const String PATH_THIRD_PARTY_ENROLLMENT_FORM = "enrollmentform-sdk/" + THIRD_PARTY_API_VERSION;

// api2coupons services
  static const String PATH_USER_COUPONS = "api2coupons/rest/user/coupons";
  static const String PATH_USER_COUPON = "api2coupons/rest/user/coupon/";
  static const String PATH_USER_COUPON_DEACTIVATE = "api2coupons/rest/user/coupon/deactivate/";
  static const String PATH_USER_COUPON_ACTIVATE = "api2coupons/rest/user/coupon/activate/";
  static const String PATH_USER_CREATE_MULTIPLE_COUPONS = "api2coupons/rest/user/createMultipleCouponsJWT";
  static const String PATH_USER_REDEMPTION_CODE = "api2coupons/rest/user/token";

  // PublisherAPI services
  static const String PUBLISHER_API_VERSION = "v2";
  static const String PATH_PUBLISHER_PROMOTIONS = "publisher/promotions/" + PUBLISHER_API_VERSION;
  static const String PATH_PUBLISHER_PROMOTION = "publisher/promotions/" + PUBLISHER_API_VERSION + "/";
  static const String PATH_PUBLISHER_CATEGORIES = "publisher/categories/" + PUBLISHER_API_VERSION;
  static const String PATH_PUBLISHER_COMPANIES = "publisher/companies/" + PUBLISHER_API_VERSION;
  static const String PATH_PUBLISHER_RETAILERS = "publisher/retailers/" + PUBLISHER_API_VERSION;
  static const String PATH_PUBLISHER_STORES = "publisher/stores/" + PUBLISHER_API_VERSION;
}

class HttpRequest extends HttpInterface {
  String url, method, httpMethod, domain;
  dynamic pathParameter;
  Map<dynamic, dynamic> properties = new Map();
  Map<dynamic, dynamic> payload = new Map();
  Map<dynamic, dynamic> arguments = new Map();
  StreamController<HttpResponse> streamController;
  Services service = new Services();

  void setDomain(String domain) {
    this.domain = domain;
  }

  void setPathParameter(dynamic pathParameter) {
    this.pathParameter = pathParameter;
  }

  void onPreExecute() {
    service.setDomain(domain);
    service.setPathParameter(pathParameter == null ? "" : pathParameter.toString());
    service.getSchema(method).forEach((schemaKey, schemaValue) {
      if (HttpInterface.PROPERTIES == schemaKey) {
        arguments.forEach((argumentsKey, argumentsValue) {
          if (schemaValue.contains(argumentsKey)) properties.putIfAbsent(argumentsKey, () => argumentsValue);
        });
      } else if (HttpInterface.PAYLOAD == schemaKey) {
        arguments.forEach((argumentsKey, argumentsValue) {
          if (schemaValue.contains(argumentsKey)) payload.putIfAbsent(argumentsKey, () => argumentsValue);
        });
      } else if (HttpInterface.HTTP_METHOD == schemaKey) {
        httpMethod = schemaValue.toString();
      } else if (HttpInterface.URL == schemaKey) {
        url = schemaValue.toString();
      }
    });
  }

  void execute() {
    onPreExecute();
    try {
      var httpClient = HttpClient();
      var httpResponse = new HttpResponse();
      var dateFormat = new DateFormat("yyyy-MM-ddTHH:mm:ss");
      var currentDate = new DateTime.now();
      var currentDateFormatted = dateFormat.format(currentDate) +
          (currentDate.timeZoneOffset.isNegative ? "-" : "+") +
          (currentDate.timeZoneOffset.abs().inHours.toString().length > 1
              ? ((currentDate.timeZoneOffset.abs().inMinutes / 60) % 12).toInt().toString()
              : "0" + ((currentDate.timeZoneOffset.abs().inMinutes / 60) % 12).toInt().toString()) +
          ((currentDate.timeZoneOffset.abs().inMinutes % 60).toString().length > 1
              ? ":" + (currentDate.timeZoneOffset.abs().inMinutes % 60).toString()
              : ":0" + (currentDate.timeZoneOffset.abs().inMinutes % 60).toString());
      if (httpMethod == HttpInterface.POST) {
        //print(url);
        httpClient.postUrl(Uri.parse(url)).then((HttpClientRequest request) {
          properties.forEach((key, value) {
            if (value != null) {
              if (key == Constants.CLIENT_API_TOKEN) {
                //special case if header auth is needed
                var secretKey = utf8.encode(properties[Constants.CLIENT_SECRET_KEY]);
                var bytes = utf8.encode(value + ":" + currentDateFormatted + ":" + json.encode(payload).toString());
                var hmacSha256 = new Hmac(sha256, secretKey); // HMAC-SHA256
                var digest = hmacSha256.convert(bytes);
                request.headers.add(HttpInterface.HEADER_AUTH_APIKEY, value);
                request.headers.add(HttpInterface.HEADER_AUTH_METHOD, HttpInterface.SIGNATURE);
                request.headers.add(HttpInterface.HEADER_AUTH_TIMESTAMP, currentDateFormatted);
                request.headers.add(HttpInterface.HEADER_AUTH_SIGNATURE, "$digest");
              } else if (key == Constants.CLIENT_SECRET_KEY) {
                //none, do not send
              } else {
                //all
                request.headers.add(key, value);
              }
            }
          });
          if (payload.length > 0) {
            request.write(json.encode(payload).toString());
          }
          /*request.headers.forEach((key, value) {
            print(key.toString() + ":" + value.toString());
          });*/
          try {
            return request.close().timeout(const Duration(seconds: HttpInterface.TIME_OUT_SC));
          }on TimeoutException catch (ex) {
            print(ex.toString());
            return null;
          } on Exception catch (ex) {
            print(ex.toString());
            return null;
          }
        }).then((HttpClientResponse response) async {
          // Process the response.
          Map<dynamic, dynamic> responseMap = new Map();
          responseMap.putIfAbsent(HttpInterface.STATUS_CODE, () => response.statusCode);
          httpResponse.statusCode = response.statusCode;
          Map<String, String> headers = new Map();
          response.headers.forEach((key, value) {
            headers.putIfAbsent('"' + key + '"', () => '"' + value[0].toString() + '"');
          });
          responseMap.putIfAbsent(HttpInterface.HEADERS, () => headers);
          String reply = await response.transform(utf8.decoder).join();
          responseMap.putIfAbsent(HttpInterface.BODY, () => reply);
          onPostExecute(HttpResponse.fromJson(responseMap));
        }).catchError((e) {
          //handle the exception the way you want, like following
          Map<dynamic, dynamic> responseMap = new Map();
          Map<String, String> headers = new Map();
          responseMap.putIfAbsent(HttpInterface.STATUS_CODE, () => HttpInterface.HTTP_REQUEST_TIMEOUT);
          responseMap.putIfAbsent(HttpInterface.HEADERS, () => headers);
          responseMap.putIfAbsent(HttpInterface.BODY, () => "'" + Constants.REQUEST_TIMEOUT + "'");
          onPostExecute(HttpResponse.fromJson(responseMap));
          print('Caught error when processing: $e');
        });
      } else if (httpMethod == HttpInterface.GET) {
        String queryParameters = "";
        try {
          var i = 0;
          payload.forEach((key, value) {
            if (value != null) {
              queryParameters =
                  (queryParameters.toString() + ((i == 0 ? "?" : "&") + (key.toString() + "=" + value.toString())));
              i += 1;
            }
          });
        } catch (ex) {
          print(ex.toString());
        }
        url = url + queryParameters;
        //print(url);
        httpClient.getUrl(Uri.parse(url)).then((HttpClientRequest request) {
          properties.forEach((key, value) {
            if (value != null) {
              //special cases if header auth is needed
              if (key == Constants.CLIENT_API_TOKEN &&
                  properties[Constants.CLIENT_SECRET_KEY] != null &&
                  properties[Constants.JWT] == null) {
                var secretKey = utf8.encode(properties[Constants.CLIENT_SECRET_KEY]);
                var bytes = utf8.encode(value + ":" + currentDateFormatted);
                var hmacSha256 = new Hmac(sha256, secretKey); // HMAC-SHA256
                var digest = hmacSha256.convert(bytes);
                request.headers.add(HttpInterface.HEADER_AUTH_APIKEY, value);
                request.headers.add(HttpInterface.HEADER_AUTH_TIMESTAMP, currentDateFormatted);
                request.headers.add(HttpInterface.HEADER_AUTH_SIGNATURE, "$digest");
                request.headers.add(HttpInterface.HEADER_AUTH_METHOD, HttpInterface.SIGNATURE);
              } else if (key == Constants.CLIENT_API_TOKEN &&
                  properties[Constants.CLIENT_SECRET_KEY] == null &&
                  properties[Constants.JWT] != null) {
                request.headers.add(HttpInterface.HEADER_AUTH_APIKEY, value);
                request.headers.add(HttpInterface.HEADER_AUTH_SIGNATURE, properties[Constants.JWT]);
                request.headers.add(HttpInterface.HEADER_AUTH_METHOD, HttpInterface.JWT);
                request.headers.add(HttpInterface.HEADER_AUTH_TIMESTAMP, currentDateFormatted);
              } else if (key == Constants.CLIENT_SECRET_KEY || key == Constants.JWT) {
                //none, do not send
              } else {
                //all
                request.headers.add(key, value);
              }
            }
          });
          /*request.headers.forEach((key, value) {
            print(key.toString() + ":" + value.toString());
          });*/
          try {
            return request.close().timeout(const Duration(seconds: HttpInterface.TIME_OUT_SC));
          }on TimeoutException catch (ex) {
            print(ex.toString());
            return null;
          } on Exception catch (ex) {
            print(ex.toString());
            return null;
          }
        }).then((HttpClientResponse response) async {
          // Process the response.
          Map<dynamic, dynamic> responseMap = new Map();
          responseMap.putIfAbsent(HttpInterface.STATUS_CODE, () => response.statusCode);
          httpResponse.statusCode = response.statusCode;
          Map<String, String> headers = new Map();
          response.headers.forEach((key, value) {
            headers.putIfAbsent('"' + key + '"', () => '"' + value[0].toString() + '"');
          });
          responseMap.putIfAbsent(HttpInterface.HEADERS, () => headers);
          String reply = await response.transform(utf8.decoder).join();
          responseMap.putIfAbsent(HttpInterface.BODY, () => reply);
          onPostExecute(HttpResponse.fromJson(responseMap));
        }).catchError((e) {
          Map<dynamic, dynamic> responseMap = new Map();
          Map<String, String> headers = new Map();
          responseMap.putIfAbsent(HttpInterface.STATUS_CODE, () => HttpInterface.HTTP_REQUEST_TIMEOUT);
          responseMap.putIfAbsent(HttpInterface.HEADERS, () => headers);
          responseMap.putIfAbsent(HttpInterface.BODY, () => "'" + Constants.REQUEST_TIMEOUT + "'");
          onPostExecute(HttpResponse.fromJson(responseMap));
          //handle the exception the way you want, like following
          print('Caught error when processing: $e');
        });
      }
    httpClient.close();
    } on Exception catch (ex) {
      print(ex.toString());
    }

  }

  void setMethod(String method) {
    this.method = method;
  }

  void setArguments(Map<dynamic, dynamic> arguments) {
    this.arguments = arguments;
  }

  void onPostExecute(HttpResponse httpResponse) {
    streamController.sink.add(httpResponse);
  }

  void setStreamController(StreamController streamController) {
    this.streamController = streamController;
  }
}

class HttpResponse {
  int statusCode;
  Map<dynamic, dynamic> headers = new Map();
  String body;

  HttpResponse();

  HttpResponse.fromJson(Map<dynamic, dynamic> json)
      : statusCode = json[HttpInterface.STATUS_CODE],
        headers = json[HttpInterface.HEADERS],
        body = json[HttpInterface.BODY];

  Map<dynamic, dynamic> toJson() => {
        HttpInterface.STATUS_CODE: statusCode,
        HttpInterface.HEADERS: headers,
        HttpInterface.BODY: body,
      };
}
